/*********************************************************************
 *
 * New idea - we don't have to use the same permutation for the whole
 * search. So each time we move down the tree we re-permute the
 * elements below the current one. Eg: if we find an element which
 * has no valid image, then we can prune that branch of the tree.
 *
 ********************************************************************/

#include "prog.h"

int
count_images (n)
     short n;
{
  /* Return the number of valid images for n
   * t is the trial element for relpres (which we overwrite)
   */
  int result;
  int rho;
  register short rp, t, np;
  register int ii, numrels;

  result = 0;
  for (t = 1; t <= SB; t++) {
    total_trials++;
    /* Start of relpres subroutine:
     * if relpres(psi,t,n) then rp=1
     * relpres(psi,t,n) is eqt to psi[n]=t; relpres(psi,n)
     * t is the test value for psi[n], relpres(psi,n-1) is true */

#include "relpres.c"
/* Uses np, rho, ii and numrels to calculate relpres(psi,t,n)
 * Returns result in rp
 */

    if (rp == 1) {
      result++;
    }
  }
  return (result);
}				/* End of count_images (n) */

int
find_good_element (n)
     short n;
{
  /* Permute pi so that n'th element has smallest no. of images
   * Returns no. of images for chosen element
   */
  int images, min_images, best_swap;
  int i;

  min_images = count_images (n);
  best_swap = n;
  for (i = n + 1; i <= SA; i++) {
    insert (i, n);		/* move element at position i to n */
    swaps++;
    setup (n + 1);
    insert (n, i);		/* undo the swap */
    images = count_images (n);
    if (images < min_images) {
      min_images = images;
      best_swap = i;
    }
  }
  /* do the best swap found: */
  if (best_swap != n) {
    insert (best_swap, n);
    swaps++;
  }
  setup (n + 1);
  return (min_images);
}				/* End of find_good_element (n) */

void
do_bush ()
{
  double count, bush_trials;
  int rho;
  double last_print, last_bush;
  int bush;
  int i, retries;
  double best_trials, best_result;
  double progress;

  register short rp, n, t, np;
  register int ii, numrels;

  trials = 0;
  total_trials = 0;
  bush_trials = 0;
  progress = 0;
  count = 0;
  n = 1;
  t = 1;
  last_print = -BUSH_PR_STEP;
  /* Don't do any bush pruning for depths up to max_depth
   * next_bush[n] is the time when the next bush prune at this depth is due */
  for (i = 1; i <= max_depth; i++) {
    next_bush[i] = MAXDOUBLE;
  }
  swaps = 0;
  if (!PRE_ANALYSIS) {
    for (i = 1; i <= SA; i++) {
      pi[i] = i;
    }
    randompi ();
  } else {
    (void) printf ("\n\n\n");
  }
  setup (SA);
  for (;;) {
    if (total_trials >= last_print + BUSH_PR_STEP) {
      last_print = total_trials;
      for (i = 1; i <= 15 && i <= SA; i++) {
	(void) printf ("%4i", psi[i]);
      }
      (void) printf ("  %.0f/%.0f\n", progress, max_depth_result);
      printinfo (1, n, bush_trials, count);
    }
    if (t <= SB) {
      total_trials++;
      /* Start of relpres subroutine:
       * if relpres(psi,t,n) then rp=1
       * relpres(psi,t,n) is eqt to psi[n]=t; relpres(psi,n)
       * t is the test value for psi[n], relpres(psi,n-1) is true */

#include "relpres.c"
/* Uses np, rho, ii and numrels to calculate relpres(psi,t,n)
 * Returns result in rp
 */

      if (rp == 1) {
	if (n == SA) {
	  count++;
	  if (DROP && (n == SA)) {
	    /* Test if current map is still relpres when the dropped
	     * relation is re-inserted. Ie test if the current map
	     * preserves the dropped relation (R).
	     * Note that we can't use rel_to_A etc. since they are not
	     * set up for relation R.
	     */
	    if (!relpres ()) {
	      psi[n] = t;
	      (void) printf ("Map found which preserves all relations except %i\n", DROP);
	      print_map ();
	      exit (0);
	    }
	  }
	  t++;
	} else {
	  /* Go deeper */
	  psi[n] = t;
	  n++;
	  t = 1;
	  /* At this point we can permute the elements at n
	   * and beyond to minimise the no. of images for psi[n]
	   * We increase the bush size until 1/BUSH_GOES of
	   * BUSH_BUDGET has been used.
	   * Then use the rest of BUSH_BUDGET to prune the bush.
	   * If any calculation returns a result of 0 without exceeding
	   * its budget, then this branch is a dead end and we can go up.
	   */
	  if ((n < (3 * SA) / 4) && (total_trials > next_bush[n])) {
	    last_bush = bush_trials;
	    /* Try to improve the current bush.
	     * First pick a suitable bush size
	     */
	    bush = 1;
	    budget = BUSH_BUDGET / BUSH_GOES;
	    for (;;) {
	      bush++;
	      best_result = calculate (n, n + bush + 1, budget);
	      bush_trials += trials;
	      if (trials > budget)
		break;
	      if ((best_result == 0) || (n + bush + 1 == SA)) {
		/* This branch has been completely fathomed */
		count += best_result;
		for (i = n; i <= n + bush + 1; i++) {
		  next_bush[i] = total_trials + (bush_trials - last_bush) * 5;
		}
		goto go_up;
	      }
	    }
	    /* Try to improve the bush from n to n+bush
	     * using up to BUSH_BUDGET effort
	     */
	    budget = BUSH_BUDGET;
	    best_result = calculate (n, n + bush, budget);
	    bush_trials += trials;
	    if (trials <= budget) {
	      budget -= trials;
	      if (PRINT_BUSH)
		(void) printf ("Bush depth = %i, size = %i, trials = %.0f, leaves = %.0f ", n, bush, trials, best_result);
	      if (best_result > 1) {
		best_trials = trials;
		for (retries = 1; retries <= BUSH_RETRIES; retries++) {
		  best_result = find_good_insert (n, n + bush, best_trials,
						  best_result, &bush_trials);
		  best_trials = trials;
		  if (inserts_done == 0)
		    break;
		}
		/* find_good_insert only sets up the arrays to current depth
		 * so we need to set them up again:
		 */
		setup (SA);
		/* Note the time when this depth should next be pruned: */
		for (i = n; i <= n + bush; i++) {
		  next_bush[i] = total_trials + (bush_trials - last_bush) * 5;
		}
	      }
	      if (PRINT_BUSH)
		(void) printf ("/ %.0f\n", best_result);
	      if (best_result == 0)
		goto go_up;
	    }
	  }
	}
      } else {
	t++;
      }				/* fi (rp == 1) */
    } else {			/* from (t <= SB) */
    go_up:
      /* progress records how many of the max_depth_result leaves
       * we have processed so far.
       */
      if (n == max_depth)
	progress++;
      n--;
      if (n < 1) {
	goto calc_od1;
      }
      t = psi[n] + 1;
    }				/* fi from (t <= SB) */
  }
calc_od1:
  (void) printf ("\n\n Finished Bush Method\n\n");
  printinfo (1, SA, total_trials, count);
}
