#include <stdio.h>
#include <sys/types.h>		/* for time() */
#include <sys/time.h>		/* to seed random() */
#include <sys/resource.h>	/* for getrusage() */
#include <values.h>		/* for MAXINT and MAXDOUBLE */
#include <malloc.h>		/* for malloc(size) */
#include <varargs.h>		/* for error function */

/* Print input data? */
int PRINT_DATA;
/* Print pcurrent permutation when other data is printed? */
int PRINT_PERM;

int ANALYSE;
int SAMPLES;			/* Initial no of samples for analyse() */
int MAXSAMPLES;			/* Max no of samples for analyse() */
int MULT;			/* Multiplier for samples when more accuracy is needed */

/* Set to 1 to get the program to choose its element order: */
int PRE_ANALYSIS;
/* These are the initial figures, so they can be quite small: */
int MAXGOES;			/* Max no of inserts tried in find_good_insert */
int BUDGET;			/* Initial budget for find_good_insert in pre-analysis */
int MAXRETRIES;			/* Max calls of find_good_insert at each depth */
int PRINTINTERVAL;		/* Interval between data prints */
/* MAXTRIES ensures a reasonable upper bound to the analysis phase time */

int BETTER;			/* 0 = minimise tree, 1 = minimise leaves */
/* Set to 1 if you want a "move" to be two insertions: */
/* may only be useful if MAXGOES and BUDGET are _very_ large */
int TWO_INSERTS;

int CALCULATE;

int BUSH;
int BUSH_BUDGET;		/* Maximum effort to spend on a single bush pruning */
int BUSH_GOES;			/* How many attempts to prune each bush */
int BUSH_RETRIES;		/* Max calls of find_good_insert for each bush */
int BUSH_PR_STEP;		/* How often to print data */
int PRINT_BUSH;			/* Print the effect of pruning each bush */
int DROP;			/* Relation to drop (0 means keep them all) */


/* Global data declarations */

int R;				/* No of relations */
int SA;				/* size of set A */
int SB;				/* size of set B */
int inserts_done;		/* set by find_good_insert */
int cum_time;			/* used by printinfo() */
int cum_utime;			/* used by printinfo() */
/* use doubles for budget values and results to avoid overflow: */
double trials;			/* set by calculate() */
double total_trials;		/* calculation budget = total trials done in pre_analysis */
double budget;			/* decreased by find_good_insert and find_good_depth_elt */

double maxtrials;
int maxgoes;
int printinterval;
int max_depth;			/* Deepest level reached by pre-analysis */
double max_depth_result;	/* No. of leaves at max_depth */

short ***RA, ***rel_to_A, ***rel_A_to, ***RB;
short *psi;
int *pi;
double *next_bush;

short *tvals;

int swaps;			/* Used to see if setup (SA) is needed */
int line_no;			/* line number in parameter file */


/* My exported procedures */

/* utils.c */
extern void error ();
extern short ***allocate_3D ();
extern void do_allocations ();
extern void readdata ();
extern void setup ();
extern void printdata ();
extern void initrandom ();
extern int rnd ();
extern void randompi ();
extern void insert ();
extern void printinfo ();
extern void next_pair ();
extern void random_pair ();
extern void print_perm ();
extern void print_map ();

/* calc.c */
extern double calculate ();
extern int relpres ();

/* analyse.c */
extern double analyse ();
extern void do_analyse ();

/* pre-analyse.c */
extern double find_good_depth_elt ();
extern double find_good_insert ();
extern void do_pre_analysis ();

/* bush.c */
extern void do_bush ();

/* read-param.c */
extern void read_param ();

/* Standard procedures used */
extern malloc_t malloc ();
extern int scanf ();
extern int printf ();
extern time_t time ();
extern long random ();
extern srandom ();
extern void exit ();
extern setlinebuf ();
extern getrusage ();
extern int fprintf ();
extern int vfprintf ();
extern double fabs ();
/* in read-param.c */
extern int sscanf ();
extern int fclose ();
/* used by getc macro: */
extern int _filbuf ();
